#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>
#include <audio.h>
#include <mappertiny.h>
#include <minifat.h>
#include <codec.h>
#include <vsNand.h>

#include <player.h>

#include <dev1000.h>

//#define VS10XX_SPEED_TEST

/*
  CHANGES from VS1000A to VS1000B/C:
  - FIX_EARSPEAKER, DISABLE_REPLAYGAIN, and PATCH_VORBIS not needed
  - LOWPOWER_PAUSE and PAUSE_TIMEOUT not needed
  - SCSI_LED2 not needed
  - voltage settings manipulation differ,
    PowerSetVoltages() takes an array pointer as parameter,
    this allows separate voltage sets for player, usb, and low-power pause
  - MassStorage() now automatically deletes/creates (tiny)mapper
    and changes voltages
  - USB Suspend is implemented, also used for low-power pause mode
    There are separate voltage settings for suspend!
    Note: a small patch is required to fully power down analog parts,
          otherwise the suspend power is around 3mA.
  - ph is initialized before nandboot, the user does not need to set it
  - SI/SO are set to GPIO mode _after_ nandboot
    -> if you replace the main loop, you must setup these yourself

  - New hooks:
    + KeyEventHandler=0x36
    + MassStorage=0x38
    + USBSuspend=0x3a (waits for changes in USBP/USBN or keys to return)
      Also used for low-power pause.
    + InitUSBDescriptors=0x3c
    + SetVolume=0x3e
  - Max Volume can be limited using the SetVolume() hook.
  - player.offDelay is the poweroff timeout value for pause mode
    (in about 5-sec resolution, default = 60 = 5mins)
  - player.maxClock sets the max clock used by the player, default=7, 3.5x
    If you increase this, increase the voltages[voltCorePlayer] also..
  - new key-event ke_ff_faster uses fast play mode for fast forward,
    and speeds up the playback as long as key is pressed.
    Use the following to return to normal play speed afterwards:
    {KEY_LONG_PRESS|KEY_4, ke_ff_faster},
    {KEY_LONG_PRESS|KEY_RELEASED, ke_ff_off},
 */


/*
  Note to MAPPERLESS:
  - this option can be used when FLASH programmer is used:
    a FAT filesystem image is written by a flash programmer starting
    from the third erase block, and bad eraseblocks are skipped
  - this disk can not be written to, thus USB should be also disabled.
  - at startup bad eraseblocks are scanned and remembered,
    and the mapper is bypassed.
  - an image creator software is available on separate request
    (or from VLSI web pages)
 */
/*
  Note to SAVE_SETTINGS:
  - with this option the play position, volume, and other settings
    are saved when the unit is turned off.
  - with MAPPERLESS the second eraseblock is used for saving
  - when eraseBlockSize >=128kB:
    the last eraseblock is used for saving,
    and the mapper receives the reduced size
    + you should not have previous content on the flash when a firmware
      with this option set is first run (when eraseBlockSize >= 128kB)
    + So, it is best to program it to a fresh board, or make certain the
      VS1000_B.RUN erases the filesystem while programming the firmware.
    + Otherwise it seems to work, but either settings are at some point
      lost when new files are written, or the settings overwrite some
      file data.
  - when eraseBlockSize <128kB:
    the second eraseblock is used for saving,
    and the mapper receives the original size
  - SAVE_NO_POSITION ignores play position but saves other settings
 */

//#define MAPPERLESS /* No mapper, no USB */
#ifndef MAPPERLESS
#define WITH_USB /* Enable USB detection if mapper in use. */
#endif/*!MAPPERLESS*/
//#define USE_CHAPTERS_CHARACTER /*first filename char groups chapters into sections */
//#define AUDIO_TESTS /* enable audio tests: KEY_5 pressed at power-on */
//#define MAX_VOLUME      3 /* Limit volume to -1.5dB of max */
//#define DEFAULT_VOLUME 34 /* -17dB */
//#define NEW_OFF_LIMIT   8 /* changes the 2-sec poweroff press time - 8=0.5sec*/
//#define LOWER_VOLTAGES    /* slightly lower voltages than default */
//#define FIRST_SONG_KEYS (KEY_1|KEY_3) /* implements goto to first song */
//#define DOUBLE_VOLUME_SPEED           /* doubles volume adjustment speed */
//#define EGG_FILES_KEYS (KEY_2|KEY_4|KEY_POWER) /*play .egg w/ key combination*/
//#define STOP_AFTER_FILES  /* goes to stop mode after all files played */
//#define POWERON_KEYWAIT 6 /* 6/16=0.375sec press required for power on */
//#define NO_WRAP           /* no wrap last->first and first->last */
//#define POWEROFF_TIMEOUT 300 /* pause -- not defined = default 300 seconds */
//#define SAVE_SETTINGS     /* Use one eraseblock for play position etc. */
//#define SAVE_NO_POSITION   /* ignore position, if SAVE_SETTINGS on */
#ifdef WITH_USB
//#define PATCH_LBAB  /* Removes 4G restriction from USB (SCSI).*/
/* For this feature you need dev1000.h in include/, libdev1000.a in lib/,
   and -ldev1000 in the vslink options in BUILD.BAT */
#endif
//#define LOOP_FILES       /* loop files, but allow prev/next etc.. */

/*experimental features*/
//#define POWER_MONITOR    /* 34 words -- quick-flash power led when low power */


extern struct FsNandPhys fsNandPhys;
extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;
extern u_int16 codecVorbis[];

void UserInterfaceIdleHook(void);

#ifdef PATCH_LBAB
#include <usblowlib.h>
#include <scsi.h>
#include <dev1000.h>
extern struct SCSIVARS {
  SCSIStageEnum State;
  SCSIStatusEnum Status;
  u_int16 *DataOutBuffer;
  u_int16 DataOutSize;
  u_int16 DataBuffer[32];
  u_int16 *DataInBuffer;
  u_int16 DataInSize;
  unsigned int BlocksLeftToSend;
  unsigned int BlocksLeftToReceive;
  u_int32 CurrentDiskSector;
  u_int32 mapperNextFlushed;
  u_int16 cswPacket[7];  /*< command reply (Command Status Wrapper) data */
  u_int32 DataTransferLength; /*< what is expected by CBW */
  s_int32 Residue; /*< difference of what is actually transmitted by SCSI */
  s_int16 DataDirection;
} SCSI;
#ifdef PATCH_TEST_UNIT_READY
void ScsiTestUnitReady(void) {
}
#endif/*PATCH_TEST_UNIT_READY*/
#endif/*PATCH_LBAB*/


#ifdef SAVE_SETTINGS
struct {
    u_int16 curMode;   /* 0 / 1  0xffff=unused 0xf0xx=valid 0x0000=deleted*/
    s_int16 curFile;   /* 0..0xffff */
    u_int16 curPos;    /* 0..0xfffe */
    u_int16 curVolume; /* -24..180 */
    u_int16 curEarSpeaker;
    u_int16 curBass;
    u_int16 reserved;
    u_int16 check; /* check-xor */
} prefs;
#endif /*SAVE_SETTINGS*/


/******************************************************************/

#ifdef SAVE_SETTINGS
u_int16 CalcCheck(register __i2 u_int16 *p) {
    register int i;
    register u_int16 chk = 0xffffU;
#if 1
    for (i=0;i<8;i++)
	chk ^= *p++;
#else /* same size.. */
    chk ^= *p++;
    chk ^= *p++;
    chk ^= *p++;
    chk ^= *p++;
    chk ^= *p++;
    chk ^= *p++;
    chk ^= *p++;
    chk ^= *p++;
#endif
    return chk;
}
#endif/*SAVE_SETTINGS*/

#ifdef SAVE_SETTINGS
void RealPowerOff(void);
void MyPowerOff(void) {
/*
  TODO: only change settings that "saved".
  TODO: do not clear+save, if the old settings match
 */
    if (cs.playTimeSeconds == 0xffffU || cs.playTimeSeconds < 5) {
	cs.playTimeSeconds = 0;
    }
    if (prefs.curMode == player.randomOn
	&& prefs.curVolume == player.volume
	&& prefs.curEarSpeaker == earSpeakerReg
	&& prefs.curBass == bassReg
#ifndef SAVE_NO_POSITION
	&& prefs.curFile == player.currentFile
	&& prefs.curPos == cs.playTimeSeconds
#endif
	) {
	/* same settings - do not save */
	return;
    }

    prefs.curMode = player.randomOn | 0xf000;
#ifdef SAVE_NO_POSITION
    prefs.curFile = 0;
    prefs.curPos = 0;
#else
    prefs.curFile = player.currentFile;
    prefs.curPos = cs.playTimeSeconds;
#endif
    prefs.curVolume = player.volume;
    prefs.curEarSpeaker = earSpeakerReg;
    prefs.curBass = bassReg;
    prefs.reserved = 0;
    prefs.check = 0;
    prefs.check = CalcCheck((void *)&prefs);

 again:
    {
	int n = ph->eraseBlockSize, erased = 0;
#ifdef MAPPERLESS
	__y u_int32 block = ph->eraseBlockSize; /* second erase block */
#else
	__y u_int32 block;
	if (n < 256) {
	    block = 256-n; /* use the last erase block before mapper */
	} else {
	    /*ph->eraseBlocks--;*/ /* already done at startup */
	    block = (u_int32)ph->eraseBlocks * n; /* use the last erase block*/
	}
#endif
	while (n--) {
	    int i = 0, touched = 0;
	    ph->Read(ph, block, 1, minifatBuffer, NULL);
	    while (i < 256) {
		if ((minifatBuffer[i] & 0xff00U) == 0xf000U) {
		    /* valid entry, make it deleted */
		    minifatBuffer[i] = 0;
		    touched = 1;
		} else if (minifatBuffer[i] == 0xffffU) {
		    /* unused entry, save settings here */
		    memcpy(minifatBuffer+i, &prefs, sizeof(prefs));
		    touched = 2;
		    break;
		}
		i += 8;
	    }
	    if (touched) {
		ph->Write(ph, block, 1, minifatBuffer, NULL);
		if (touched == 2) {
		    break; /* Done.. */
		}
	    }
	    block++;
	    if (n == 0 && erased == 0) {
		/* Was last chance.. Erase and start over. */
		n = ph->eraseBlockSize;
#ifdef MAPPERLESS
		block = ph->eraseBlockSize;
#else
		if (n < 256) {
		    block = 256-n; /* use the last erase block before mapper */
		} else {
		    block = (u_int32)ph->eraseBlocks * n;
		}
#endif
		ph->Erase(ph, block);
		erased = 1;
	    }
	}
    }
    RealPowerOff();
}
#endif/*SAVE_SETTINGS*/



#if defined(SAVE_SETTINGS) && !defined(SAVE_NO_POSITION)
s_int16 CsOutput(struct CodecServices *cs, s_int16 *data, s_int16 n);
s_int16 MyCsOutput(struct CodecServices *cs, s_int16 *data, s_int16 n) {
  if (prefs.curMode & 0xf000U) {
      prefs.curMode &= ~0xf000U;
      if (prefs.curPos) {
	  cs->goTo = prefs.curPos;
      }
  }
  return CsOutput(cs, data, n); /* handles load checking */
}
#endif /*defined(SAVE_SETTINGS) && !defined(SAVE_NO_POSITION)*/


enum keyEvent2 {
  // end of default events
#ifdef FIRST_SONG_KEYS
  ke_first = 64, /* goto first song */
#endif
#ifdef EGG_FILES_KEYS
  ke_egg = 65,   /* play .egg files */
#endif
#ifdef USE_CHAPTERS_CHARACTER
  ke_prev_section = 66,
  ke_next_section = 67,
#endif
};
const struct KeyMapping myKeyMap[] = {
  {KEY_POWER, ke_pauseToggle},
  {KEY_LONG_PRESS|KEY_POWER, ke_powerOff},

#ifdef DOUBLE_VOLUME_SPEED /* double the volume adjustment speed */
  {KEY_1, ke_volumeDown2},
  {KEY_LONG_PRESS|KEY_1, ke_volumeDown2},
  {KEY_2, ke_volumeUp2},
  {KEY_LONG_PRESS|KEY_2, ke_volumeUp2},
#else
  {KEY_1, ke_volumeDown},
  {KEY_LONG_PRESS|KEY_1, ke_volumeDown},
  {KEY_2, ke_volumeUp},
  {KEY_LONG_PRESS|KEY_2, ke_volumeUp},
#endif
  {KEY_3, ke_previous},
  {KEY_LONG_PRESS|KEY_3, ke_rewind},
  {KEY_4, ke_next},
#if 0
  {KEY_LONG_PRESS|KEY_4, ke_forward},
#else
  {KEY_LONG_PRESS|KEY_4, ke_ff_faster},
  {KEY_LONG_PRESS|KEY_RELEASED, ke_ff_off},
#endif
  {KEY_5, ke_earSpeaker},
  {KEY_LONG_PRESS|KEY_5, (s_int16)(ke_randomToggleNewSong|KEY_LONG_ONESHOT)},
#ifdef FIRST_SONG_KEYS
  {FIRST_SONG_KEYS, ke_first}, /* goto first song */
#endif
#ifdef EGG_FILES_KEYS
  {EGG_FILES_KEYS, ke_egg},   /* play .egg files */
#endif
  {0, ke_null}
};
#ifdef AUDIO_TESTS
const struct KeyMapping myEmptyMap[] = {
  {KEY_LONG_PRESS|KEY_POWER, ke_powerOff},
  {0, ke_null}
};
#endif

/*
  Play/Pause + Power On  - POWERBUTTON
  Stop       + Power Off   KEY_2
  Volume Toggle            KEY_1
  Next                     KEY_4
  Prev                     KEY_3
 */
#ifdef USE_CHAPTERS_CHARACTER
u_int16 prevChapter, nextChapter;
#endif

#if defined(FIRST_SONG_KEYS) || defined(EGG_FILES_KEYS) || defined(DOUBLE_VOLUME_SPEED) || defined(NO_WRAP) || defined(STOP_AFTER_FILES)
void MyKeyEventHandler(enum keyEvent event) {
  switch(event) {
#ifdef USE_CHAPTERS_CHARACTER
  case ke_prev_section:
      prevChapter = minifatInfo.fileName[0];
      if (player.currentFile && cs.playTimeSeconds < 5) {
	  /* jump to the beginning of the previous group */
	  player.nextFile = player.currentFile - 1;
      } else {
	  /* jump to the beginning of this group */
	  player.nextFile = player.currentFile;
      }
      cs.cancel = 1;
      player.pauseOn = 0;
      break;
  case ke_next_section:
      nextChapter = minifatInfo.fileName[0];
      cs.cancel = 1;
      player.pauseOn = 0;
      break;
#endif
#ifdef FIRST_SONG_KEYS
  case ke_first: /* goto first song */
      player.nextFile = 0;
      cs.cancel = 1;
      player.pauseOn = 0;
      break;
#endif

#ifdef EGG_FILES_KEYS
  case ke_egg:   /* go and play .egg files */
  {
      static const u_int32 eggSupportedFiles[] = {
	  FAT_MKID('E','G','G'), 0
      };
      u_int16 newTotal;
      minifatInfo.supportedSuffixes = &eggSupportedFiles[0];
      newTotal = OpenFile(0xffffU);
      if (newTotal) {
	  supportedFiles = &eggSupportedFiles[0]; /* only "EGG" */
	  player.totalFiles = newTotal;
	  cs.cancel = 1;
	  player.nextFile = 0;
	  player.pauseOn = 0;
      } else {
	  minifatInfo.supportedSuffixes = supportedFiles;
      }
      break;
  }
#endif /*EGG_FILES_KEYS*/

#if defined(STOP_AFTER_FILES) || defined(NO_WRAP)
  case ke_previous:
#ifdef STOP_AFTER_FILES
      if (player.nextFile == player.currentFile) {
	  /* pause mode is stop mode */
      } else
#endif/*STOP_AFTER_FILES*/
	  if (cs.playTimeSeconds < 5
#ifdef NO_WRAP
		 && player.currentFile
#endif
	  ) {
	  player.nextFile = player.currentFile - 1;
      } else {
	  player.nextFile = player.currentFile;
      }
      cs.cancel = 1; /* jump to the new file */
      player.pauseOn = 0; /* normally done by cs.cancel, needed now */
      player.nextStep = -1; /* skip unsupported files backward */
      break;
#endif/*defined(STOP_AFTER_FILES) || defined(NO_WRAP)*/

#if defined(NO_WRAP) && !defined(STOP_AFTER_FILES)
  case ke_next:
      if (player.nextFile < player.totalFiles) {
	  cs.cancel = 1; /* jump to the new file */
	  player.pauseOn = 0; /* normally done by cs.cancel, needed now */
	  player.nextStep = 1; /* skip unsupported files forward */
      }
      break;
#endif

#if defined(STOP_AFTER_FILES)
  case ke_pauseToggle:
      if (player.pauseOn && player.nextFile == player.currentFile) {
	  /* pause mode is stop mode */
	  player.nextFile = player.currentFile + 1;
      }
      player.pauseOn = player.pauseOn ? 0 : 1;
      break;
#endif

  default:
      /* execute the default actions (earspeaker and volume) */
      RealKeyEventHandler(event);
      break;
  }
}
#endif /*defined(FIRST_SONG_KEYS) || defined(EGG_FILES_KEYS) || defined(DOUBLE_VOLUME_SPEED) || defined(NO_WRAP) || defined(STOP_AFTER_FILES)*/

extern u_int16 keyCheck;
void MyUserInterfaceIdleHook(void) {
    if (uiTrigger) {
	register u_int16 check = 0;
	register u_int16 keyNew;
	uiTrigger = 0;

#ifndef NEW_OFF_LIMIT
	KeyScan();
#else
	if (USBIsAttached()) {
	    player.pauseOn = 0;
	    cs.cancel = 1;
	}
	keyNew = ReadGPIO() & 31;
	if (USEX(SCI_STATUS) & SCISTF_REGU_POWERBUT) {
	    keyNew |= KEY_POWER;
	}
	if (keyNew == keyOld) {
	    ++keyOldTime;
	    if (keyOld & KEY_POWER) { /* note: other keys can be pressed */
		if (keyOldTime < NEW_OFF_LIMIT) {
		    check = 0;
		} else if (keyOldTime == NEW_OFF_LIMIT) { /* OFF_LIMIT seconds */
		    check = KEY_POWER | KEY_LONG_PRESS;
		}
	    } else if (keyOldTime >= SHORT_LIMIT) {
		/* long press */
		check = keyOld | KEY_LONG_PRESS;
	    }
	} else {
	    if ((s_int16)keyOldTime >= 0 && keyOldTime < SHORT_LIMIT) {
		/* Short press: was pressed < 0.5 seconds */
		check = keyOld;
	    } else {
		check = KEY_RELEASED | KEY_LONG_PRESS; /*any long press released!*/
	    }
	    keyOld = keyNew;
	    keyOldTime = 0;
	}
	if (check) {
	    register const struct KeyMapping *p = currentKeyMap;
	    keyCheck = check;
	    while (p->key) {
		if (check == p->key &&
		    (p->event > 0 || keyOldTime <= SHORT_LIMIT)) {
		    KeyEventHandler(p->event & 0x7fff);
		}
		p++;
	    }
	}
#endif

	{
	    register u_int16 leds = LED1;
	    if (player.pauseOn && (uiTime & 0x06) != 0)
		leds = 0; /* power led flashes when in pause mode */
#ifdef POWER_MONITOR
	    {
		static int lowCount = 0;
		if (USEX(SCI_STATUS) & SCISTF_REGU_POWERLOW) {
		    lowCount++;
		} else if (lowCount) {
		    lowCount--;
		}
		if (lowCount > 5 && (uiTime & 1) == 0) {
		    leds = 0;
		}
	    }
#endif
	    if (player.randomOn)
		leds ^= LED2;
	    PERIP(GPIO1_ODATA) = (PERIP(GPIO1_ODATA) & ~(LED1|LED2)) | leds;
	}
    }
}



#ifdef AUDIO_TESTS
auto s_int16 Sine(register __a0 u_int16 ph);
s_int16 ISine(u_int16 ph) {
    if (ph & 7) {
	register u_int16 f = (ph << 13);
	return (s_int16)(((s_int32)Sine(ph) * (f^0xffffU) +
			  (s_int32)Sine(ph+8) * f) >> 16);
    }
    return Sine(ph);
}

void AudioTests(void) {
    extClock4KHz = 12288/4;
    clockX = 7; /* 3.5x */
    SetRate(48000U);
    bassReg = earSpeakerReg = 0;
    volumeReg = 0;
    SetVolume();

    /* Fix digital volume */
#define AP_IMD
    {
	register u_int32 phs = 0, add = 0;
	__y u_int16 ph3, ph5, ph7;
	register s_int16 *p = (s_int16 *)minifatBuffer;
	u_int16 wait = 0, mode = 0, ph2, add2;
	while (1) {
	    if (keyCheck == KEY_3) {
		keyCheck = 0;
		add += 0x5d00000UL; /* ~1kHz step */
	    }
	    if (keyCheck == KEY_4) {
		keyCheck = 0;
		add -= 0x5d00000UL; /* ~1kHz step */
	    }
	    if (keyCheck == KEY_1) { /* vol down */
		keyCheck = 0;
		if (mode == 2) {
		    volumeReg ^= 0x00feU;
		    SetVolume();
		}
	    }
	    if (keyCheck == KEY_2) {
		keyCheck = 0;
		if (mode == 0) {
		    if (USEX(DAC_VOL) == 0xb0b0) {
			USEX(DAC_VOL) = 0x0000;
		    } else {
			USEX(DAC_VOL) += 0x1010;
		    }
		} else if (mode == 2) {
		    volumeReg ^= 0xfe00U;
		    SetVolume();
		}
	    }
	    if (keyCheck == KEY_POWER) {
		/* next test */
		keyCheck = 0;
		if (++mode > 3)
		    mode = 0;
		if (mode == 0) {
		    /* frequency sweep */
		    extClock4KHz = 12288/4;
		    clockX = 8; /* 4.0x */
		    SetRate(48000U);
		    bassReg = earSpeakerReg = 0;
		    volumeReg = 0;
		    SetVolume();
		}
		if (mode == 1) {
		    /* amplitude sweep */
		    extClock4KHz = 12000/4;
		    SetRate(44100U);
		    volumeReg = 0x8080;
		    SetVolume();
		    add = 0x05d00000UL;
		}
		if (mode == 2) {
		    /* sin / crosstalk */
		    volumeReg = 0x00fe;
		    add = 0x05d00000UL * 10; /* Start at 10kHz */
		    SetVolume();
		}
		if (mode == 3) {
		    /* IMD 0.75% 60Hz 0.25% 7kHz */
		    extClock4KHz = 12288/4;
		    SetRate(48000U);
		    volumeReg = 0x0000;
		    SetVolume();
#ifdef AP_IMD
		    /* 3kHz pulse, ~15kHz sin */
		    add2 = 4096;
		    //add  = 20490L*65536;
		    add  = 20480L*65536;
		    ph2 = ph3 = ph5 = ph7 = 0;
#else
		    add = 0x00590000UL; /* 60Hz */
		    add2 = 0x28b0; /* 7kHz */
#endif
		}
	    }
	    if (mode == 3) {
#ifdef AP_IMD
		s_int16 tmp;

		tmp = ((s_int32)ISine(ph2)*16384U +
		       (s_int32)ISine(ph3)*5461U +
		       (s_int32)ISine(ph5)*3277U +
		       (s_int32)ISine(ph7)*2341U)>>14;

		p[0] = p[1] = //tmp;
		    (52428U * (s_int32)tmp + (s_int32)ISine(phs>>16) * 13107) >> 16;
		ph3 += add2*3;
		ph5 += add2*5;
		ph7 += add2*7;
#else
		p[0] = p[1] =
		    ((s_int32)ISine(phs>>16) * 49152U +
		     (s_int32)ISine(ph2)*16384) >> 16;
#endif
		ph2 += add2;
	    } else if (mode == 2) {
		if ((volumeReg & 0xff00U) == 0xfe00U) {
		    p[0] = 0;
		} else {
		    p[0] = ISine(phs>>16);
		}
		if ((volumeReg & 0xff) == 0xfe) {
		    p[1] = 0;
		} else {
		    p[1] = ISine(phs>>16);
		}
	    } else {
		p[0] = p[1] = ISine(phs>>16);
	    }
	    p += 2;
	    phs += add;

	    if (p >= &minifatBuffer[2*16]) {
		p = (s_int16 *)minifatBuffer;
		AudioOutputSamples(p, 16);

		if (mode == 0) {
		    if (add < 0x01000000UL) {
			add += 0x40;
		    } else if (add < 0x10000000UL) {
			add += 0x400;
		    } else {
			add += 0x4000;
		    }
		    if (add >= 0x80000000UL) {
			add = 0;
		    }
		} else if (mode == 1) {
		    wait++;
		    if (wait > 4000) {
			wait = 0;
			if (volumeReg > 0) {
			    volumeReg -= 0x0101;
			    SetVolume();
			}
		    }
		} else if (mode == 2) {
#ifdef AUTO_CROSS
		    wait++;
		    if (wait > (s_int16)(441000/16)) {
			wait = 0;
			volumeReg = (volumeReg<<8)|(volumeReg>>8);
			SetVolume();
		    }
#endif
		}
	    }
	}
    }
}
#endif/*AUDIO_TESTS*/

#ifdef MAPPERLESS
#define MAX_BADS 16
s_int16 numOfBads;
u_int32 bads[MAX_BADS];
auto u_int16 MyReadDiskSector(register __i0 u_int16 *buffer,
			      register __reg_a u_int32 sector) {
    __y u_int32 sect = sector + 2*ph->eraseBlockSize;
    int i = 0;
    while (i < numOfBads) {
	if (sect < bads[i])
	    break;
	sect += ph->eraseBlockSize;
	i++;
    }
#if 0
    puthex(sect);
    puts("=read");
#endif
    return (u_int16)ph->Read(ph, sect, 1, buffer, NULL);
}
#endif

#ifdef MAX_VOLUME
auto void MySetVolume(void) {
#if 0 /* If Left/Right can have different volumes. */
    if ((volumeReg & 0x00ff) < MAX_VOLUME)
	volumeReg = (volumeReg & 0xff00U) | MAX_VOLUME;
    if ((volumeReg >> 8) < MAX_VOLUME)
	volumeReg = (volumeReg & 0x00ffU) | (MAX_VOLUME<<8);
#else
    if (volumeReg < 0x0101*MAX_VOLUME)
	volumeReg = 0x0101*MAX_VOLUME;
#endif
    RealSetVolume();
}
#endif/*MAX_VOLUME*/

#ifdef LOOP_FILES
/* For seamless playing of the same file in a loop. These are
   not normally used, so we don't bother including codecvorbis.h . */
auto s_int16 CodVorbisGoTo(register struct CodecVorbis *cv,
			   register u_int16 seconds);
struct Codec *CodVorbisCreate(void);
extern u_int16 ogg[];
extern s_int16 vFirstFrame;
#endif

s_int16 UserParameter(void);
s_int16 UserConfig(void);


#ifdef VS10XX_SPEED_TEST
#include <dev1000.h>
u_int16 Mp3ReadReg(register __c0 u_int16 reg) {
    /* assert /CS (drop FSYNC) */
    PERIP(GPIO0_ODATA) &= ~GPIO0_CS1;
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE1;
    SpiDelay(100);
    SpiSendReceive(0x0300 | reg);
    reg = SpiSendReceive(0xffff);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE1;
    PERIP(GPIO0_ODATA) |= GPIO0_CS1;
    SpiDelay(100);
    return reg;
}
void Mp3WriteReg(register __c0 u_int16 reg, register __c1 val) {
    /* assert /CS (drop FSYNC) */
    PERIP(GPIO0_ODATA) &= ~GPIO0_CS1;
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE1;
    SpiDelay(100);
    SpiSendReceive(0x0200 | reg);
    SpiSendReceive(val);
    SpiDelay(100);
    PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE1;
    PERIP(GPIO0_ODATA) |= GPIO0_CS1;
    SpiDelay(100);
}
#endif


void main(void) {
#ifdef VS10XX_SPEED_TEST
    /* VS10xx SCI speed test.. */
    {
	int myClk;
	u_int16 vsClk;

	/* CS1 = xCS
	   ALE = xDCS
	*/
		    /* VS1000 VS1003 div  res
		       3.5x   1.0x   8 *2 ok   CLKI/4.57
		       3.5x   1.5x   6 *2 ok   CLKI/5.14
		       3.5x   2.0x   5 *2 ok   CLKI/5.71
		       3.5x   2.5x   4 *2 ok   CLKI/5.71
		       3.5x   3.0x   4 *2 ok   CLKI/6.85 (5.14 fails)
		       3.5x   3.5x   3 *2 ok   CLKI/6
		       3.5x   4.0x   3 *2 ok   CLKI/6.85 (4.57 fails)
		       3.5x   4.5x   3 *2 ok   CLKI/7.71 (5.14 fails)
	   4.5x   4.5x   4 *2 ok   CLKI/8
	   4.5x   4.5x   3 *2 fail CLKI/6!!
	   4.5x   4.0x   4 *2 ok   CLKI/7.11
	   4.5x   4.0x   3 *2 fail (CLKI/5.333)
	   4.5x   3.5x   4 *2 ok   CLKI/6.222
	   4.5x   3.0x   5 *2 ok   CLKI/6.666
	   4.5x   2.5x   6 *2 ok   CLKI/6.666
	   4.5x   2.5x   5 *2 ok   CLKI/5.555
	   4.5x   2.0x   6 *2 ok   CLKI/5.333

	   4.0x   3.5x   4 *2 ok   CLKI/7
	   4.0x   4.0x   3 *2 fail CLKI/6!!
	   4.0x   4.0x   4 *2 ok   CLKI/8
	   4.0x   3.0x   3 *2 fail
	   4.0x   3.0x   4 *2 ok   CLKI/6
	   4.0x   2.5x   5 *2 ok   CLKI/6.25
	   4.0x   2.0x   6 *2 ok   CLKI/6
	   4.0x   1.5x   7 *2 ok   CLKI/5.25
	   4.0x   1.5x   6 *2 fail (CLKI/4.5)
	   4.0x   1.0x   10*2 ok   CLKI/5
	   3.0x   1.0x   7 *2 fail (CLKI/4.666)
	   3.0x   1.0x   8 *2 ok   CLKI/5.333
	   3.0x   1.5x   5 *2 ok   CLKI/5
	   3.0x   2.0x   4 *2 ok   CLKI/5.333
	   3.0x   2.5x   4 *2 ok   CLKI/6.666
	   3.0x   2.5x   3 *2 fail (CLKI/5)
	   3.0x   3.0x   3 *2 ok   CLKI/6
	   3.0x   3.5x   3 *2 ok   CLKI/7
	   3.0x   3.5x   2 *2 fail (CLKI/4.666)
	   3.0x   4.0x   3 *2 ok   CLKI/8
	   3.0x   4.0x   2 *2 fail (CLKI/5.333)
	   3.0x   4.5x   3 *2 ok   CLKI/9
	   3.0x   4.5x   2 *2 fail CLKI/6!!!
	*/
	voltages[voltCorePlayer] = 31;
	PowerSetVoltages(&voltages[voltCorePlayer]);

	PERIP(GPIO1_MODE) |= (LED1|LED2);

	PERIP(SPI0_CONFIG) = SPI_CF_MASTER | SPI_CF_DLEN16 | SPI_CF_FSIDLE1;
	PERIP(SPI0_FSYNC) = 0xffff;
	PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (64-1);
	PERIP(GPIO0_ODATA) |= GPIO0_ALE | GPIO0_CS1;
	PERIP(GPIO0_MODE) &= ~(GPIO0_ALE | GPIO0_CS1);
	PERIP(GPIO0_DDR)  |= (GPIO0_ALE | GPIO0_CS1);

	for (myClk = 2; myClk < 10; myClk++) {
	    clockX = myClk;
	    SetRate(44100U);
	    puthex(clockX);
	    puts("=clockx");
	    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (64-1);
	    Mp3ReadReg(0x00);
	    vsClk = 0;
	    do {
		int div = 1;
		long loops;
		PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (64-1);
		Mp3ReadReg(0x00);
		Mp3WriteReg(0, 0x4c00);
		Mp3ReadReg(0x00);
		Mp3WriteReg(0x03, vsClk);
		puthex(Mp3ReadReg(0x3));
		puts("=CLOCKF");
	    again:
		for (loops = 12000; loops > 0 ; loops--) {
		    register u_int16 f;
		    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (div-1);
		    Mp3WriteReg(0x0c, 0xaaaa);
		    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (20-1);
		    f = Mp3ReadReg(0x0c);
		    if (f != 0xaaaa) {
#if 0
		    puthex(f);
		    puts("=aaaa");
#endif
		    break;
		    }
		    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (div-1);
		    Mp3WriteReg(0x0c, 0x5555);
		    PERIP(SPI0_CLKCONFIG) = SPI_CC_CLKDIV * (20-1);
		    f = Mp3ReadReg(0x0c);
		    if (f != 0x5555) {
#if 0
		    puthex(f);
		    puts("=5555");
#endif
		    break;
		    }
		}
		if (loops) {
		    div++;
#if 0
		    puthex(div);
		    puts("=div");
#endif
		    goto again;
		}
		puthex(div);
		puts("=div");
		vsClk += 0x2000;
	    } while (vsClk);
	}
    }
#endif

#ifdef USE_CHAPTERS_CHARACTER
    prevChapter = 0;
    nextChapter = 0;
#endif
#ifdef POWEROFF_TIMEOUT
    player.offDelay = POWEROFF_TIMEOUT/5; /* in about 5-sec resolution, default = 60 = 5mins */
#endif
    //ph = &fsNandPhys.p;
#ifdef DEFAULT_VOLUME
    player.volume = DEFAULT_VOLUME-24
#ifdef MAX_VOLUME
	-MAX_VOLUME
#endif
	;
#endif

#ifdef MAPPERLESS
    /* Setup -- Scan bad erase blocks. */
    {
	register u_int16 i;
	numOfBads = 0;
	for (i=1; i<ph->eraseBlocks; i++) {
	    /* read both spares */
	    ph->Read(ph, (u_int32)i * ph->eraseBlockSize, 2, NULL, minifatBuffer);
	    if ((minifatBuffer[0/2 + 2] & 0xff) != 0xff ||
		(minifatBuffer[16/2+ 2] & 0xff) != 0xff) {
		/* bad */
		bads[numOfBads++] = (u_int32)i * ph->eraseBlockSize;
#if 0
		puthex(i);
		puts("=bad");
#endif
		if (numOfBads == MAX_BADS)
		    break;
	    }
	}
    }
    /* Replace read function -- bypass mapper. */
    SetHookFunction((u_int16)ReadDiskSector, MyReadDiskSector);
#endif/*MAPPERLESS*/


#ifdef SAVE_SETTINGS
    prefs.curMode = 0;
    {
	int n = ph->eraseBlockSize;
#ifdef MAPPERLESS
	__y u_int32 block = ph->eraseBlockSize; /* second eraseblock */
#else
	__y u_int32 block;
	if (n < 256) {
	    block = 256-n; /* use the last erase block before mapper */
	} else {
	    ph->eraseBlocks--; /* Reserve one erase block for our settings */
	    block = (u_int32)ph->eraseBlocks * n; /* use the last erase block*/
	}
#endif
	while (n--) {
	    register int i = 0;
	    ph->Read(ph, block, 1, minifatBuffer, NULL);
	    while (i < 256) {
		if (i < 256 && (minifatBuffer[i] & 0xff00U) == 0xf000U) {
		    /* valid entry */
		    memcpy(&prefs, minifatBuffer+i, sizeof(prefs));
		    break;
		}
		i += 8;
	    }
	    block++;
	}
    }
    if (CalcCheck((void *)&prefs)) {
	prefs.curMode = 0; /* Invalid entry, discard */
    }
#endif


    /** Set default voltage startup values and registers, then load */
#ifdef LOWER_VOLTAGES
    //voltages[voltCoreSuspend] = 14; /* 1.8V */
    voltages[voltCoreSuspend] = 12; /* 1.75V */
    voltages[voltIoSuspend]   = 13; /* 2.4V */
    voltages[voltAnaSuspend]  =  0; /* 2.36V */

    voltages[voltCorePlayer] = 22; /*2.2V */
    voltages[voltIoPlayer]   = 18; /*2.7V */
    //voltages[voltAnaPlayer]  = 6; /* 2.6V Note: lower than recommended! */
    PowerSetVoltages(&voltages[voltCorePlayer]);
#endif/*LOWER_VOLTAGES*/

    /* Set the leds after nand-boot! */
    /* We set LED2 here too to prevent the display to mess it up later.. */
    PERIP(GPIO1_ODATA) |=  LED1|LED2;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

#if defined(FIRST_SONG_KEYS) || defined(EGG_FILES_KEYS) || defined(DOUBLE_VOLUME_SPEED)
    currentKeyMap = myKeyMap;
#endif
#if defined(FIRST_SONG_KEYS) || defined(EGG_FILES_KEYS) || defined(DOUBLE_VOLUME_SPEED) || defined(NO_WRAP) || defined(STOP_AFTER_FILES) || defined(NEW_OFF_LIMIT)
    SetHookFunction((u_int16)IdleHook, MyUserInterfaceIdleHook);
#endif

    player.volumeOffset = -24;
    player.pauseOn = 0;
    //player.randomOn = 0;

#ifdef AUDIO_TESTS
    keyCheck = 0; /* polled key functions */
    if ((ReadGPIO() & 31) == KEY_5) {
	/* Feature button pressed when powerup */
	currentKeyMap = myEmptyMap;
	AudioTests();
    }
#endif

#ifdef MAX_VOLUME
    SetHookFunction((u_int16)SetVolume, MySetVolume);
#endif
#ifdef PATCH_LBAB
    /* Increases the allowed disk size from 4GB to 2TB.
       Reference to PatchMSCPacketFromPC will link the function from
       libdev1000.a. We need to implement ScsiTestUnitReady(). */
    SetHookFunction((u_int16)MSCPacketFromPC, PatchMSCPacketFromPC);
#endif/*PATCH_LBAB*/

#ifdef POWERON_KEYWAIT
    /* Wait until powerbutton up */
 more:
    while ((USEX(SCI_STATUS) & SCISTF_REGU_POWERBUT)) {
	if (timeCount > (POWERON_KEYWAIT*TIMER_TICKS/16)) {
	    PERIP(GPIO1_ODATA) |= LED1;
	    break;
	}
	if ((timeCount & 0xc)==0) {
	    PERIP(GPIO1_ODATA) |= LED1;
	} else {
	    PERIP(GPIO1_ODATA) &= ~LED1;
	}
    }
    /* If < 1.5sec, power off */
    if (timeCount < (POWERON_KEYWAIT*TIMER_TICKS/16)) {
	PERIP(GPIO1_ODATA) &= ~LED1;
	/* wait at least 1 sec before poweroff
	   to make certain power is properly ON. */
	while (timeCount < TIMER_TICKS) {
	    if ((USEX(SCI_STATUS) & SCISTF_REGU_POWERBUT))
		goto more;
	    Sleep();
	}
	PowerOff(); /* no settings changed */
    }
    PERIP(GPIO1_ODATA) |= LED1;
#endif

#if defined(SAVE_SETTINGS)
    /* Set hook after POWERON_KEYWAIT so we only save settings after
       a real power-on + power-off. */
    SetHookFunction((u_int16)PowerOff, MyPowerOff);
#if !defined(SAVE_NO_POSITION)
    /* If play position needs to be restored, do it in codec output function */
    cs.Output = MyCsOutput;
#endif
#endif

#if defined(FIRST_SONG_KEYS) || defined(EGG_FILES_KEYS) || defined(DOUBLE_VOLUME_SPEED) || defined(NO_WRAP) || defined(STOP_AFTER_FILES)
    SetHookFunction((u_int16)KeyEventHandler, MyKeyEventHandler);
#endif

    keyOld = KEY_POWER;
    keyOldTime = -32767; /* ignores the first release of KEY_POWER */

    /* replicate the whole main loop code here.. */
#ifndef MAPPERLESS
    map = FsMapTnCreate(ph, 0); /* tiny mapper */
#endif/*!MAPPERLESS*/

    PlayerVolume();
    while (1) {

#ifdef WITH_USB
	if (USBIsAttached()) {
#ifdef LOWER_VOLTAGES
	    /* This is one way to use different voltages for USB suspend
	       and low-power pause mode.
	       Note: the pause mode would like a little patch to reduce
	       the USB suspend current from about 3mA to around 0.6mA.
	    */
	    voltages[voltCoreSuspend] = 14; /* 1.8V */
	    voltages[voltIoSuspend]   = 31; /* 3.3V */
	    voltages[voltAnaSuspend]  = 30; /* 3.6V */
#endif
	    MassStorage();
#ifdef LOWER_VOLTAGES
	    voltages[voltCoreSuspend] = 10; /* 1.7V */
	    voltages[voltIoSuspend]   =  8; /* 2.2V */
	    voltages[voltAnaSuspend]  =  0;
#endif
	}
#endif
	if (InitFileSystem() == 0) {
	    minifatInfo.supportedSuffixes = supportedFiles;
	    player.totalFiles = OpenFile(0xffffU);
	    if (player.totalFiles == 0)
		goto noFSnorFiles;

	    player.nextStep = 1;
	    player.nextFile = 0;

#ifdef SAVE_SETTINGS
	    if (prefs.curMode & 0xf000U) {
#ifdef SAVE_NO_POSITION
		prefs.curMode &= ~0xf000U; /* no jump */
#else
		if ((player.randomOn = (prefs.curMode & 1)))
		    prefs.curMode &= ~0xf000U; /* if random play -> no jump */
		player.nextFile = prefs.curFile;
#endif
		player.volume = prefs.curVolume;
		earSpeakerReg = prefs.curEarSpeaker;
		bassReg = prefs.curBass;
	    }
#endif
	    while (1) {
		if (player.randomOn) {
		    register s_int16 nxt;
		retoss:
		    nxt = rand() % player.totalFiles;
		    if (nxt == player.currentFile && player.totalFiles > 1)
			goto retoss;
		    player.currentFile = nxt;
		} else {
		    player.currentFile = player.nextFile;
		}
		if (player.currentFile < 0)
		    player.currentFile += player.totalFiles;
#ifdef STOP_AFTER_FILES
		if (player.currentFile >= player.totalFiles) {
		    /* goto pause=stop mode if played last file to the end. */
		    keyOld     = KEY_POWER;
		    keyOldTime = SHORT_LIMIT-2;
		    player.nextFile = player.currentFile = 0;
		} else {
		    player.nextFile = player.currentFile + 1;
		}
#else
		if (player.currentFile >= player.totalFiles) {
		    player.currentFile -= player.totalFiles;
		}
		player.nextFile = player.currentFile + 1;
#endif

		if (OpenFile(player.currentFile) < 0) {

		    //player.volumeOffset = -12; /*default volume offset*/
		    //PlayerVolume();
		    player.ffCount = 0;
		    cs.cancel = 0;
		    cs.goTo = -1;
		    cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
		    cs.fastForward = 1; /* reset play speed to normal */

		    {
#if 0 /*The implements skipping unsupported files backwards. */
			register s_int16 oldStep = player.nextStep;
#endif
			register s_int16 ret;
#ifdef LOOP_FILES
			LoadCheck(NULL, 0); /*higher clock*/
			if ((cod = CodVorbisCreate())) {
			    const char *eStr;
			    ret = cod->Decode(cod, &cs, &eStr);
			    /* if played to the end, start over */
			    while (cs.fileLeft < 5) {
				LoadCheck(NULL, 0); /*higher clock*/
				CodVorbisGoTo(cod, 0);
				ogg[1]/*headerTypeFlag*/ = 0;
				ogg[14]/*cont*/ = 0;
				vFirstFrame = 1;
				cs.playTimeSeconds = cs.playTimeSamples = 0;
				ret = cod->Decode(cod, &cs, &eStr);
			    }
			    cod->Delete(cod);
			}
#else
			ret = PlayCurrentFile();
#endif

#if 0 /*This implements skipping unsupported files backwards. */
			/* If unsupported, keep skipping */
			if (ret == ceFormatNotFound)
			    player.nextFile = player.currentFile + player.nextStep;
			/* If supported file found, restore play direction */
			if (ret == ceOk && oldStep == -1)
			    player.nextStep = 1;
#endif
		    }
		} else {
		    player.nextFile = 0;
		}
#ifdef USE_CHAPTERS_CHARACTER
		if (prevChapter) {
		    register int cur = player.nextFile;
		    if (cur != player.currentFile && OpenFile(cur) < 0) {
			prevChapter = minifatInfo.fileName[0];
		    }
		    --cur;
		    while (cur >= 0 && OpenFile(cur) < 0) {
			if ((minifatInfo.fileName[0] ^ prevChapter) &0xff00U) {
			    /* different first char found */
			    cur++;
			    break;
			}
			cur--;
		    }
		    if (cur < 0)
			cur = 0;
		    player.nextFile = cur;
		    prevChapter = 0;
		}
		if (nextChapter) {
		    register int cur = player.currentFile+1;
		    while (cur < player.totalFiles && OpenFile(cur) < 0) {
			if ((minifatInfo.fileName[0] ^ nextChapter) &0xff00U) {
			    /* different first char found */
			    break;
			}
			cur++;
		    }
		    player.nextFile = cur;
		    nextChapter = 0;
		}
#endif
#ifdef WITH_USB
		if (USBIsAttached()) {
		    break;
		}
#endif
	    }
	} else {
	noFSnorFiles:
	    LoadCheck(&cs, 32); /* decrease or increase clock */
	    memset(tmpBuf, 0, sizeof(tmpBuf));
	    AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	    /* When no samples fit, calls the user interface
	       -- handle volume control and power-off */
	}
    }
}
